<?php

/**
 * @file
 * BS Base drush integration offering various theme management function.
 */

use Drush\Log\LogLevel;
use Drupal\Component\Serialization\Yaml;
use Drupal\Core\Extension\Discovery\RecursiveExtensionFilterIterator;
use Drupal\Core\Site\Settings;
use Symfony\Component\Routing\Generator\UrlGenerator;

/**
 * Implements hook_drush_command().
 */
function bs_base_drush_command() {
  $items = [];

  $items['bs-theme-create'] = [
    'description' => 'Create a new bs_base compatible child theme.',
    'arguments' => [
      'parent-machine-name' => 'Parent theme machine name.',
      'child-machine-name' => 'Child theme machine name.',
      'child-name' => 'Child theme name.',
      'child-description' => 'Child theme description.',
    ],
    'required-arguments' => TRUE,
    'examples' => [
      "drush bs-ts bs_bootstrap custom_theme 'Custom Theme' 'Custom Theme description'" => "Create a new child custom_theme in the standard themes/custom/custom_thmee folder.",
    ],
    'bootstrap' => DRUSH_BOOTSTRAP_DRUPAL_ROOT,
    'aliases' => ['bs-tc'],
  ];

  $items['bs-theme-update'] = [
    'description' => 'Update existing bs_base compatible child theme.',
    'arguments' => [
      ' theme-machine-name' => 'Theme machine name.',
    ],
    'required-arguments' => TRUE,
    'examples' => [
      'drush bs-tu custom_theme' => 'Update custom_theme theme.',
    ],
    'bootstrap' => DRUSH_BOOTSTRAP_DRUPAL_ROOT,
    'aliases' => ['bs-tu'],
  ];

  $items['bs-theme-build'] = [
    'description' => 'Download theme package.json dependencies with yarn and build theme CSS files.',
    'arguments' => [
      'theme-machine-name' => 'Theme machine name.',
    ],
    'required-arguments' => TRUE,
    'examples' => [
      "drush bs-tb custom_theme" => "Download custom_theme build dependencies and build all CSS files.",
    ],
    'aliases' => ['bs-tb'],
  ];

  return $items;
}

/**
 * Callback for the bs-theme-create drush command.
 *
 * Create a new child theme based on passed parent theme.
 *
 * @param string $parent_machine_name
 *   Parent theme machine name.
 * @param string $child_machine_name
 *   Child theme machine name.
 * @param string $child_name
 *   Child name.
 * @param string $child_description
 *   Child description name.
 */
function drush_bs_base_bs_theme_create($parent_machine_name, $child_machine_name, $child_name, $child_description) {
  // Verify that the child machine name contains no disallowed characters.
  if (preg_match('@[^a-z0-9_]+@', $child_machine_name)) {
    drush_log('The machine-readable name "' . $child_machine_name . '" must contain only lowercase letters, numbers, and hyphens.', LogLevel::ERROR);
    return;
  }

  drush_log("Starting $child_machine_name theme creation", LogLevel::OK);

  // Parent theme should exist.
  $parent_path = _bs_base_drupal_get_theme_path($parent_machine_name);
  if (empty($parent_path)) {
    drush_log('Parent theme does not exist.', LogLevel::ERROR);
    return;
  }

  // Child theme should not exist.
  if (!empty($child_path = _bs_base_drupal_get_theme_path($child_machine_name))) {
    drush_log("Child theme already exist on $child_path file system.", LogLevel::ERROR);
    return;
  }

  // Create child theme directory.
  $child_path = 'themes/custom/' . $child_machine_name;
  if (!mkdir($child_path, 0755, TRUE)) {
    drush_log("Failed to create child theme directory on $child_path path.", LogLevel::ERROR);
    return;
  }

  $options = [
    'parent_machine_name' => $parent_machine_name,
    'parent_path' => $parent_path,
    'child_machine_name' => $child_machine_name,
    'child_path' => $child_path,
    'child_name' => $child_name,
    'child_description' => $child_description,
  ];

  // Copy files from parent and change/apply text changes to labels.
  _bs_base_copy_theme_files($options);

  // Replace text in copied files.
  _bs_base_reconfigure_theme_files($options);

  // Generate some files from the scratch.
  _bs_base_generate_file('theme-options.yml', $options);
  _bs_base_generate_file("config/schema/{$child_machine_name}.schema.yml", $options);
  _bs_base_generate_file('gulp-options.yml', $options);
  _bs_base_generate_file($child_machine_name . '.info.yml', $options);
  _bs_base_generate_file($child_machine_name . '.libraries.yml', $options);
  _bs_base_generate_file('README.md', $options);

  // Rebuild themes static cache because new theme is created.
  _bs_base_drupal_theme_list_info(TRUE);

  // Make sure we are on latest parent theme versions.
  $update_functions = _bs_base_get_update_hooks($child_machine_name);
  if (!empty($update_functions)) {
    $bs_versions = [];
    foreach ($update_functions as $theme_name => $theme_updates) {
      // Get last update.
      end($theme_updates['functions']);
      $last_function = key($theme_updates['functions']);
      $bs_versions['bs_versions.' . $theme_name] = (int) $last_function;
    }

    $all_themes = _bs_base_drupal_theme_list_info();
    _bs_base_set_yml_value($all_themes[$child_machine_name]->pathname, $bs_versions, TRUE);
  }

  // Update and flatten SASS files.
  _bs_base_update_sass_files($child_machine_name);

  // Rebuild CSS files.
  drush_bs_base_bs_theme_build($child_machine_name);
}

/**
 * Callback for the bs-theme-update command.
 *
 * Updating existing bs_base child theme.
 *
 * @param string $target_machine_name
 *   Theme machine name.
 */
function drush_bs_base_bs_theme_update($target_machine_name) {
  drush_log("Updating a $target_machine_name theme", LogLevel::OK);

  $target_path = _bs_base_drupal_get_theme_path($target_machine_name);
  if (empty($target_path)) {
    drush_log('Target theme does not exist.', LogLevel::ERROR);
    return;
  }

  $parent_themes = _bs_base_get_parent_themes($target_machine_name);
  if (empty($parent_themes)) {
    drush_log('Parent themes are missing.', LogLevel::ERROR);
    return;
  }

  // Run update hooks.
  _bs_base_theme_run_update_hooks($target_machine_name);
  drush_print();

  $all_themes = _bs_base_drupal_theme_list_info();
  $first_parent_machine_name = _bs_base_drupal_get_parent_theme_name($target_machine_name);

  _bs_base_update_sass_files($target_machine_name);

  // Check for any new or removed CSS library in parent theme and update
  // libraries-override section.
  $parent_theme_libraries_override = _bs_base_generate_libraries_override($first_parent_machine_name);
  $target_info_array = $all_themes[$target_machine_name]->info;
  $target_theme_libraries_override = $target_info_array['libraries-override'];
  // Keep only the libraries from target that are not from parents.
  foreach ($target_theme_libraries_override as $library_key => $library_value) {
    // We have it in parent already.
    if (isset($parent_theme_libraries_override[$library_key])) {
      unset($target_theme_libraries_override[$library_key]);
    }
    // We do not have it in parent but it does belong to parent themes. We
    // assume that library was removed from parent theme or that it is in parent
    // parents. In this case we will remove it from here.
    elseif (_bs_base_library_key_coming_from_parents($library_key, $target_machine_name)) {
      unset($target_theme_libraries_override[$library_key]);
    }
  }
  // We start from parent generated libraries override.
  $new_libraries_override = array_merge($parent_theme_libraries_override, $target_theme_libraries_override);
  // Update info file with new libraries override.
  $info_content = file_get_contents($all_themes[$target_machine_name]->pathname);
  $info_content = preg_replace('/^libraries-override:\R(^ +.+\R)+/m', Yaml::encode(['libraries-override' => $new_libraries_override]), $info_content);
  file_put_contents($all_themes[$target_machine_name]->pathname, $info_content);

  // Rebuild CSS files.
  drush_bs_base_bs_theme_build($target_machine_name);
}

/**
 * Callback for the bs-theme-build drush command.
 *
 * Build child theme assets.
 *
 * @param string $theme_machine_name
 *   Theme machine name.
 */
function drush_bs_base_bs_theme_build($theme_machine_name) {
  drush_log("Building CSS asset for a $theme_machine_name theme", LogLevel::OK);

  $target_path = _bs_base_drupal_get_theme_path($theme_machine_name);
  if (empty($target_path)) {
    drush_log("Target theme {$theme_machine_name} does not exist.", LogLevel::ERROR);
    return;
  }

  _bs_base_drush_build_css($target_path);
}

/* Helper functions */

/**
 * Returns array with unique lines.
 *
 * This will eliminate all duplicate lines in array but it will also compare
 * commented lines with not commented and eliminate that duplicates also. For
 * example:
 *
 * array(
 *   '@import "bs_bootstrap/sass/components/partials/alert";',
 *   '//@import "bs_bootstrap/sass/components/partials/alert";',
 * )
 *
 * This two values are considered duplicates also and the result will be
 *
 * array(
 *   '//@import "bs_bootstrap/sass/components/partials/alert";',
 * )
 *
 * @param array $lines
 *   Array of lines.
 *
 * @return array
 *   Array of unique lines.
 */
function _bs_base_array_unique_lines(array $lines) {
  $unique_lines = array();
  $matches = [];

  foreach ($lines as $key => $line) {
    // Empty line we just add and continue.
    if ($line === "\n") {
      $unique_lines[] = $line;
      continue;
    }

    // Get import path parts.
    preg_match("#(//|/\*)?\s*@import\s*['\"]?(.*)/(.*)(\.scss)?['\"]#", $line, $matches);

    // If import path is not in unique lines then lets add it. We check partial
    // name with `_` and without `_` character on the start.
    $pattern = "#@import\s*['\"]{$matches[2]}/_?{$matches[3]}(\.scss)?['\"]#";
    $res = preg_grep($pattern, $unique_lines);
    if (empty($res)) {
      // Compare line comment variation.
      // @todo - this comparator is weak and will only work for simple comment
      // case. If more is needed we will need to use some kind of regular
      // expression.
      if (strpos($line, '//') !== 0) {
        $comment_line = '//' . $line;
        if (in_array($comment_line, $lines)) {
          // If it exist we will save commented version.
          $unique_lines[] = $comment_line;
          continue;
        }
      }
      $unique_lines[] = $line;
    }
    else {
      // When we have a match we will override previous line because match can
      // happen with comment and without comment, and new line is always the
      // strongest.
      $unique_lines[key($res)] = $line;
    }
  }

  return $unique_lines;
}

/**
 * Copy general theme files from parent theme to child theme.
 *
 * @param array $options
 *   Array of options having next keys:
 *   - parent_machine_name
 *   - parent_path
 *   - child_path
 *   - child_name
 *   - child_description.
 */
function _bs_base_copy_theme_files(array $options) {
  // Files and folders which we will copy if they exist in parent theme.
  $files = [
    'config',
    'config/install',
    'config/install/' . $options['parent_machine_name'] . '.settings.yml',
    'config/schema',
    'templates',
    'templates/README.md',
    '.gitignore',
    '.browserslistrc',
    'favicon.ico',
    'gulp-tasks.js',
    'gulpfile.js',
    'logo.svg',
    'package.json',
    'screenshot.png',
    'yarn.lock',
  ];

  foreach ($files as $file) {
    _bs_base_copy_file($file, $options);
  }
}

/**
 * Copy file from parent theme to child theme.
 *
 * If parent file is a directory then it will be created.
 *
 * Target sub-directories of a file will be created automatically if they do not
 * exists.
 *
 * @param string $file
 *   File name.
 * @param array $options
 *   Array of options having next keys:
 *   - parent_machine_name
 *   - parent_path
 *   - child_machine_name
 *   - child_path.
 *
 * @return bool
 *   TRUE if the file was copied, FALSE otherwise.
 */
function _bs_base_copy_file($file, array $options) {
  // Do file rename if needed.
  $new_file_name = str_replace($options['parent_machine_name'], $options['child_machine_name'], $file);

  // If source file does not exist, log a warning and return.
  if (!file_exists($options['parent_path'] . '/' . $file)) {
    drush_log("Source file $file in {$options['parent_path']} does not exist, can not copy.", LogLevel::WARNING);
    return FALSE;
  }

  // If target file exist just return.
  if (file_exists($options['child_path'] . '/' . $new_file_name)) {
    return FALSE;
  }

  // If parent file is directory then create directory.
  if (is_dir($options['parent_path'] . '/' . $file)) {
    return mkdir($options['child_path'] . '/' . $new_file_name, 0755);
  }

  // In the case when file name has directory in it make sure that all
  // sub-directories exists in target before doing actual file copy.
  if (!_bs_base_ensure_directory($new_file_name, $options['child_path'])) {
    return FALSE;
  }

  // Copy file from parent theme folder to child theme folder.
  if (!copy($options['parent_path'] . '/' . $file, $options['child_path'] . '/' . $new_file_name)) {
    drush_log("Failed to copy $file file from {$options['parent_path']} to {$options['child_path']}.", LogLevel::ERROR);
    return FALSE;
  }

  return TRUE;
}

/**
 * Finds all the base themes for the specified theme.
 *
 * @param array $themes
 *   An array of available themes.
 * @param string $theme
 *   The name of the theme whose base we are looking for.
 *
 * @return array
 *   Returns an array of all of the theme's ancestors including specified theme.
 */
function _bs_base_drupal_get_base_themes(array $themes, $theme) {
  $base_themes = [$theme => $themes[$theme]->info['name']];

  if (!empty($themes[$theme]->info['base theme'])) {
    return _bs_base_drupal_get_base_themes($themes, $themes[$theme]->info['base theme']) + $base_themes;
  }

  return $base_themes;
}

/**
 * Returns the first parent theme of passed child theme.
 *
 * @param string $theme_name
 *   The name of the child theme whose first parent theme we are looking for.
 *
 * @return string|NULL
 *   Returns a theme machine name of first parent theme or NULL if parent does
 *   not exist.
 */
function _bs_base_drupal_get_parent_theme_name($theme_name) {
  $themes_info = _bs_base_drupal_theme_list_info();
  $parent_themes = _bs_base_drupal_get_base_themes($themes_info, $theme_name);

  end($parent_themes);
  if (!prev($parent_themes)) {
    return NULL;
  }

  return key($parent_themes);
}

/**
 * Returns the path to a Drupal theme.
 *
 * @param string $name
 *   Theme machine name.
 *
 * @return string
 *   The path to the requested theme or an empty string if the item is not
 *   found.
 */
function _bs_base_drupal_get_theme_path($name) {
  $scan = _bs_base_drupal_scan('theme');

  if (isset($scan[$name])) {
    return $scan[$name]->subpath;
  }

  return '';
}

/**
 * Discovers available extensions of a given type.
 *
 * For an explanation of how this work see ExtensionDiscovery::scan().
 *
 * @param string $type
 *   The extension type to search for. One of 'profile', 'module', 'theme', or
 *   'theme_engine'.
 * @param bool $reset
 *   Reset internal cache.
 *
 * @return stdClass[]
 *   An associative array of stdClass objects, keyed by extension name.
 */
function _bs_base_drupal_scan($type, $reset = FALSE) {
  static $processed_files = NULL;

  if (!$reset && !is_null($processed_files)) {
    return $processed_files;
  }

  $files = [];
  static $searchdirs = ['core', 'sites/all', ''];
  foreach ($searchdirs as $dir) {
    $scan_res = _bs_base_drupal_scan_directory($dir);
    // Only return extensions of the requested type.
    if (isset($scan_res[$type])) {
      $files += $scan_res[$type];
    }
  }

  // Duplicate files found in later search directories take precedence over
  // earlier ones; they replace the extension in the existing $files array.
  $processed_files = [];
  foreach ($files as $file) {
    $processed_files[basename($file->pathname, '.info.yml')] = $file;
  }

  return $processed_files;
}

/**
 * Recursively scans a base directory for the extensions it contains.
 *
 * For an explanation of how this work @see ExtensionDiscovery::scanDirectory().
 *
 * @param string $dir
 *   A relative base directory path to scan, without trailing slash.
 *
 * @return stdClass[]
 *   An associative array of stdClass objects, keyed by extension name.
 */
function _bs_base_drupal_scan_directory($dir) {
  $files = [];

  $dir_prefix = ($dir == '' ? '' : "$dir/");
  $absolute_dir = ($dir == '' ? DRUPAL_ROOT : DRUPAL_ROOT . "/$dir");

  if (!is_dir($absolute_dir)) {
    return $files;
  }

  $flags = \FilesystemIterator::UNIX_PATHS;
  $flags |= \FilesystemIterator::SKIP_DOTS;
  $flags |= \FilesystemIterator::FOLLOW_SYMLINKS;
  $flags |= \FilesystemIterator::CURRENT_AS_SELF;
  $directory_iterator = new \RecursiveDirectoryIterator($absolute_dir, $flags);

  $ignore_directories = Settings::get('file_scan_ignore_directories', []);

  $filter = new RecursiveExtensionFilterIterator($directory_iterator, $ignore_directories);

  $iterator = new \RecursiveIteratorIterator($filter,
    \RecursiveIteratorIterator::LEAVES_ONLY,
    // Suppress filesystem errors in case a directory cannot be accessed.
    \RecursiveIteratorIterator::CATCH_GET_CHILD
  );

  foreach ($iterator as $key => $fileinfo) {
    // All extension names in Drupal have to be valid PHP function names due
    // to the module hook architecture.
    if (!preg_match('/^[a-zA-Z_\x7f-\xff][a-zA-Z0-9_\x7f-\xff]*$/', $fileinfo->getBasename('.info.yml'))) {
      continue;
    }

    // Determine extension type from info file.
    $type = FALSE;
    /** @var SplFileObject $file */
    $file = $fileinfo->openFile('r');
    while (!$type && !$file->eof()) {
      preg_match('@^type:\s*(\'|")?(\w+)\1?\s*$@', $file->fgets(), $matches);
      if (isset($matches[2])) {
        $type = $matches[2];
      }
    }
    if (empty($type)) {
      continue;
    }
    $name = $fileinfo->getBasename('.info.yml');
    $pathname = $dir_prefix . $fileinfo->getSubPathname();

    // Determine whether the extension has a main extension file.
    // For theme engines, the file extension is .engine.
    if ($type == 'theme_engine') {
      $filename = $name . '.engine';
    }
    // For profiles/modules/themes, it is the extension type.
    else {
      $filename = $name . '.' . $type;
    }
    if (!file_exists(DRUPAL_ROOT . '/' . dirname($pathname) . '/' . $filename)) {
      $filename = NULL;
    }

    $extension = new stdClass();
    $extension->type = $type;
    $extension->pathname = $pathname;
    $extension->filename = $filename;
    $extension->subpath = $fileinfo->getSubPath();
    // Extension parent folder path.
    $extension->parentPath = substr($fileinfo->getPath(), 0, -(strlen($name) + 1));
    $extension->origin = $dir;

    $files[$type][$key] = $extension;
  }

  return $files;
}

/**
 * Get information's for all themes.
 *
 * @param bool $reset
 *   Reset internal cache.
 *
 * @return array
 *   Array holding themes information's.
 */
function _bs_base_drupal_theme_list_info($reset = FALSE) {
  static $themes = [];

  if (!$reset && !empty($themes)) {
    return $themes;
  }

  $themes = _bs_base_drupal_scan('theme', $reset);
  foreach ($themes as $theme_name => $theme) {
    $themes[$theme_name]->info = Yaml::decode(file_get_contents($theme->pathname));
  }

  return $themes;
}

/**
 * Finds all parent themes for the specified theme.
 *
 * @param string $theme_machine_name
 *   The machine name of the theme whose parent themes we are looking for.
 *
 * @return array
 *   Returns an array of all of the parent themes.
 */
function _bs_base_get_parent_themes($theme_machine_name) {
  $all_themes = _bs_base_drupal_theme_list_info();
  $parent_themes = _bs_base_drupal_get_base_themes($all_themes, $theme_machine_name);
  array_pop($parent_themes);
  return $parent_themes;
}

/**
 * Get theme information.
 *
 * @param string $theme_machine_name
 *   Theme machine name.
 *
 * @return mixed|null
 *   Theme info object or NULL if theme does not exist.
 */
function _bs_base_get_theme_info($theme_machine_name) {
  $all_themes = _bs_base_drupal_theme_list_info();
  return isset($all_themes[$theme_machine_name]) ? $all_themes[$theme_machine_name] : NULL;
}

/**
 * Get all child themes for passed parent_theme.
 *
 * @param string $parent_theme
 *   Machine name of parent theme.
 *
 * @return array
 *   Array of all child themes machine names. Empty array if child themes does
 *   not exist.
 */
function _bs_base_find_child_themes($parent_theme) {
  $child_themes = [];

  $themes = _bs_base_drupal_theme_list_info();
  foreach ($themes as $theme => $theme_info) {
    if ($theme === $parent_theme) {
      continue;
    }
    $parent_themes = _bs_base_drupal_get_base_themes($themes, $theme);
    if (isset($parent_themes[$parent_theme])) {
      $child_themes[$theme] = $theme;
    }
  }

  return $child_themes;
}

/**
 * Check that library is coming from theme parent themes or bs_lib module.
 *
 * @param string $library_key
 *   Library key.
 * @param string $theme_machine_name
 *   Theme machine name.
 *
 * @return bool
 *   TRUE if library is coming from parents, FALSE other way.
 */
function _bs_base_library_key_coming_from_parents($library_key, $theme_machine_name) {
  // Add bs_lib module to the parents array.
  $parent_themes = ['bs_lib' => 'bs_lib'] + _bs_base_get_parent_themes($theme_machine_name);
  foreach ($parent_themes as $parent_theme_machine_name => $parent_theme_label) {
    if (strpos($library_key, $parent_theme_machine_name) === 0) {
      return TRUE;
    }
  }

  return FALSE;
}

/**
 * Replace text in theme files so all configurations are correct.
 *
 * @param array $options
 *   Array of options having next keys:
 *   - parent_machine_name
 *   - parent_path
 *   - child_path
 *   - child_name
 *   - child_description.
 */
function _bs_base_reconfigure_theme_files(array $options) {
  // Yaml value replaces.
  $yaml_files = [
    'config/install/' . $options['child_machine_name'] . '.settings.yml' => [
      'logo.path' => $options['child_path'] . '/logo.svg',
      'favicon.path' => $options['child_path'] . '/favicon.ico',
    ],
  ];
  foreach ($yaml_files as $file => $value) {
    _bs_base_set_yml_value($options['child_path'] . '/' . $file, $value);
  }

  // Regexp string replacements.
  $regexp_files = [
    'gulpfile.js' => [
      $options['parent_machine_name'] => $options['child_machine_name'],
    ],
    'package.json' => [
      "\"name\":\s*\"${options['parent_machine_name']}\"" => "\"name\": \"${options['child_machine_name']}\"",
      "\"description\":\s*\".*\"" => "\"description\": \"${options['child_description']}\"",
    ],
  ];
  foreach ($regexp_files as $file => $regexps) {
    _bs_base_regexp_file($options['child_path'] . '/' . $file, $regexps);
  }
}

/**
 * Get theme update functions.
 *
 * @param string $target_machine_name
 *   Target theme machine name.
 *
 * @return array
 *   Array of update functions.
 */
function _bs_base_get_update_hooks($target_machine_name) {
  $update_functions = [];
  $all_themes = _bs_base_drupal_theme_list_info();

  // Get current parents version information for this theme. If not defined this
  // means that we will run all existing update functions.
  $bs_versions = isset($all_themes[$target_machine_name]->info['bs_versions']) ? $all_themes[$target_machine_name]->info['bs_versions'] : [];

  // Get parent themes.
  $parent_themes = _bs_base_get_parent_themes($target_machine_name);

  // Cycle through all parent themes and find update functions.
  foreach (array_keys($parent_themes) as $parent_theme) {
    // If install file exist get all update functions from it.
    $install_filepath = $all_themes[$parent_theme]->subpath . '/' . $parent_theme . '.bs_base.install';

    if (file_exists($install_filepath)) {
      $content = file_get_contents($install_filepath);

      // Find update functions and first comment line.
      if (preg_match_all('/\s\*\s(.*?)\n\s\*\/\nfunction\s' . $parent_theme . '_bs_update_(\d{4})/', $content, $matches)) {
        $functions = array_combine($matches[2], $matches[1]);

        // Filter update functions that were run in the past.
        if (isset($bs_versions[$parent_theme])) {
          $parent_theme_version = isset($bs_versions[$parent_theme]) ? (int) $bs_versions[$parent_theme] : 0;
          $functions = array_filter($functions, function ($version) use ($parent_theme_version) {
            return (int) $version > $parent_theme_version;
          }, ARRAY_FILTER_USE_KEY);
        }

        if (!empty($functions)) {
          $update_functions[$parent_theme] = [
            'file' => $install_filepath,
            'functions' => $functions,
          ];
        }
      }
    }
  }

  return $update_functions;
}

/**
 * Run theme update functions.
 *
 * @param string $target_machine_name
 *   Target theme machine name.
 */
function _bs_base_theme_run_update_hooks($target_machine_name) {
  $update_functions = _bs_base_get_update_hooks($target_machine_name);
  if (empty($update_functions)) {
    drush_log(dt("No theme updates required."), LogLevel::SUCCESS);
    return;
  }

  // Print a list of pending updates for this module and get confirmation.
  drush_print(dt('The following updates are pending:'));
  drush_print();

  foreach ($update_functions as $theme_name => $theme_updates) {
    drush_print($theme_name . ' theme : ');
    foreach ($theme_updates['functions'] as $version => $description) {
      drush_print($version . ' -   ' . strip_tags($description), 2);
    }
  }
  drush_print();

  if (!drush_confirm(dt('Do you wish to run all pending updates?'))) {
    return drush_user_abort();
  }

  drush_print();
  drush_print(dt('Running next updates:'));

  // Load install files and execute update functions.
  $bs_versions = [];
  foreach ($update_functions as $theme_name => $theme_updates) {
    include_once $theme_updates['file'];
    foreach ($theme_updates['functions'] as $version => $description) {
      $update_function = $theme_name . '_bs_update_' . $version;
      drush_print($version . ' -   ' . strip_tags($description), 2);
      $update_function($target_machine_name);
    }

    // Update theme info bs_versions.
    $bs_versions['bs_versions.' . $theme_name] = (int) $version;
  }

  // Update info file with latest versions.
  $all_themes = _bs_base_drupal_theme_list_info();
  _bs_base_set_yml_value($all_themes[$target_machine_name]->pathname, $bs_versions, TRUE);
}

/**
 * Execute yarn install and gulp sass in provided path.
 *
 * @param string $path
 *   Path to theme folder.
 */
function _bs_base_drush_build_css($path) {
  // Install npm packages and execute gulp sass compilation.
  drush_log("Installing any missing package and rebuilding sass files", LogLevel::OK);
  exec('cd ' . $path . ' && yarn install');
  drush_log("Rebuilding CSS files", LogLevel::OK);
  exec('cd ' . $path . ' && yarn gulp clean:css && yarn gulp sass');
}

/**
 * Ensure that theme filename has all directories.
 *
 * @param string $filename
 *   Filename with optional subdirectories in path that we ensure they exist.
 * @param string $path
 *   Base path in which we will check $filename subdirectories.
 *
 * @return bool
 *   TRUE on success, FALSE on error.
 */
function _bs_base_ensure_directory($filename, $path) {
  // In the case when file name has directory in it make sure that all
  // sub-directories exists in target before doing actual file copy.
  $dirname = dirname($filename);
  if ($dirname && !is_dir($path . '/' . $dirname)) {
    if (!mkdir($path . '/' . $dirname, 0755, TRUE)) {
      drush_log("Failed to create copy target directory $dirname.", LogLevel::ERROR);
      return FALSE;
    }
  }
  return TRUE;
}

/**
 * Finds all files that match a given mask in a given directory.
 *
 * For additional information see file_scan_directory().
 *
 * @param $dir
 *   The base directory or URI to scan, without trailing slash.
 * @param $mask
 *   The preg_match() regular expression for files to be included.
 * @param $options
 *   An associative array of additional options, with the following elements:
 *   - 'nomask': The preg_match() regular expression for files to be excluded.
 *     Defaults to the 'file_scan_ignore_directories' setting.
 *   - 'callback': The callback function to call for each match. There is no
 *     default callback.
 *   - 'recurse': When TRUE, the directory scan will recurse the entire tree
 *     starting at the provided directory. Defaults to TRUE.
 *   - 'key': The key to be used for the returned associative array of files.
 *     Possible values are 'uri', for the file's URI; 'filename', for the
 *     basename of the file; and 'name' for the name of the file without the
 *     extension. Defaults to 'uri'.
 *   - 'min_depth': Minimum depth of directories to return files from. Defaults
 *     to 0.
 * @param $depth
 *   The current depth of recursion. This parameter is only used internally and
 *   should not be passed in.
 *
 * @return
 *   An associative array (keyed on the chosen key) of objects with 'uri',
 *   'filename', and 'name' properties corresponding to the matched files.
 */
function _bs_base_file_scan_directory($dir, $mask, $options = [], $depth = 0) {
  // Merge in defaults.
  $options += [
    'callback' => 0,
    'recurse' => TRUE,
    'key' => 'uri',
    'min_depth' => 0,
  ];
  // Normalize $dir only once.
  if ($depth == 0) {
    $dir_has_slash = (substr($dir, -1) === '/');
  }

  $options['key'] = in_array($options['key'], ['uri', 'filename', 'name']) ? $options['key'] : 'uri';
  $files = [];
  // Avoid warnings when opendir does not have the permissions to open a
  // directory.
  if (is_dir($dir)) {
    if ($handle = @opendir($dir)) {
      while (FALSE !== ($filename = readdir($handle))) {
        // Skip this file if it matches the nomask or starts with a dot.
        if ($filename[0] != '.'
          && !(isset($options['nomask']) && preg_match($options['nomask'], $filename))
          && !(!empty($default_nomask) && preg_match($default_nomask, $filename))
        ) {
          if ($depth == 0 && $dir_has_slash) {
            $uri = "$dir$filename";
          }
          else {
            $uri = "$dir/$filename";
          }
          if ($options['recurse'] && is_dir($uri)) {
            // Give priority to files in this folder by merging them in after
            // any subdirectory files.
            $files = array_merge(_bs_base_file_scan_directory($uri, $mask, $options, $depth + 1), $files);
          }
          elseif ($depth >= $options['min_depth'] && preg_match($mask, $filename)) {
            // Always use this match over anything already set in $files with
            // the same $options['key'].
            $file = new stdClass();
            $file->uri = $uri;
            $file->filename = $filename;
            $file->name = pathinfo($filename, PATHINFO_FILENAME);
            $key = $options['key'];
            $files[$file->$key] = $file;
            if ($options['callback']) {
              $options['callback']($uri);
            }
          }
        }
      }

      closedir($handle);
    }
  }

  return $files;
}

/**
 * Flatten all parent theme @import directives in a SASS file.
 *
 * @param object $target_sass_file
 *   Target SASS file object.
 * @param string $target_machine_name
 *   Target theme machine name.
 * @param array $current_themes
 *   Array of current themes.
 * @param array $parent_themes_sass_files
 *   Array of all SASS files from parent theme.
 * @param int $depth
 *   Current recursion depth, internally used.
 *
 * @return array
 *   Returns array of flattened SASS @import directives.
 */
function _bs_base_flatten_sass_file_imports($target_sass_file, $target_machine_name, array $current_themes, array $parent_themes_sass_files, $depth = 0) {
  $flatten_lines = [];

  $lines = file($target_sass_file->uri);
  if ($lines === FALSE) {
    drush_log("Failed to open {$target_sass_file->uri} file.", LogLevel::ERROR);
    return [];
  }

  // For the depth 0 even when we are already flattened (no direct imports to
  // parent main SASS file) we still need to check parent theme SASS file in
  // case that parent files have additional changes that are not in target
  // files.
  $check_parent = FALSE;
  if ($depth === 0) {
    $check_parent = TRUE;
  }
  else {
    // For the depth greater then 0 we will check parent theme only if it has
    // direct import.
    $parent_imports = [];
    foreach (array_keys($parent_themes_sass_files) as $parent_theme) {
      $parent_imports = array_merge($parent_imports, preg_grep("#^@import\s*['\"]{$parent_theme}((?!/partials/).)*['\"]#", $lines));
    }
    if (!empty($parent_imports)) {
      $check_parent = TRUE;
    }
  }
  if ($check_parent) {
    $additional_lines = [];
    end($parent_themes_sass_files);
    $first_parent_theme = key($parent_themes_sass_files);
    foreach ($parent_themes_sass_files[$first_parent_theme] as $file_path => $file) {
      if ($target_sass_file->filepath === $file->filepath) {
        // Remove 'file-name.scss' part from a file path.
        $sass_import_path = substr($file->filepath, 0, -strlen($file->filename));
        // This patter will pick all possible import variation we need, with
        // partial sign and without, with .scss extension and without, commented
        // or not commented:
        // For example all next cases are considered positive for us:
        //   @import "parent_theme/sass/theme/print";
        //   @import "parent_theme/sass/theme/print.scss";
        //   @import "parent_theme/sass/theme/_print";
        //   @import "parent_theme/sass/theme/_print.scss";
        //   //@import "parent_theme/sass/theme/print";
        //   // and all other // or /* combinations.
        $pattern = "#(?://|/\*)?\s*@import\s*['\"]{$first_parent_theme}/{$sass_import_path}_?{$file->name}(\.scss)?['\"]#";

        // Add additional parent imports only if they already don't exist in
        // current lines.
        if (empty(preg_grep($pattern, $lines))) {
          $rest_of_parent_themes = $parent_themes_sass_files;
          array_pop($rest_of_parent_themes);
          $additional_flattened_files = _bs_base_flatten_sass_file_imports($file, $first_parent_theme, $current_themes, $rest_of_parent_themes, $depth + 1);
          $additional_lines = _bs_base_array_unique_lines(array_merge($additional_lines, $additional_flattened_files));
        }
        break;
      }
    }
    // If there are new additional lines from parent themes we need to check
    // them also. This will produce duplicate lines so we need to take that into
    // consideration.
    if (!empty($additional_lines)) {
      // If first line is init the lets add parent imports after it.
      if (strpos($lines[0], '@import "init"') === 0) {
        // Eliminate duplicate lines and comment duplicate lines.
        $lines = _bs_base_array_unique_lines(array_merge([array_shift($lines)], $additional_lines, $lines));
      }
      else {
        $lines = _bs_base_array_unique_lines(array_merge($additional_lines, $lines));
      }
    }
  }

  foreach ($lines as $line) {
    if ($depth === 0) {
      // If it does not start with @import just add it and move on.
      if (strpos($line, '@import') !== 0) {
        $flatten_lines[] = $line;
        continue;
      }
    }
    // Skip empty lines, and commented lines from parent theme SASS files that
    // are NOT @import directives.
    elseif ($depth > 0 && (empty($line) || (strpos($line, '//') === 0 && strpos($line, '@import') === FALSE))) {
      continue;
    }

    // Find and flatten @import directives.
    if (preg_match("#^(//|/\*)?\s*(@import\s+['\"])([a-zA-Z0-9_]+)(.*?)['\"]#", $line, $matches)) {
      $sass_file_commented = !empty($matches[1]);
      $sass_file_context = $matches[3];
      $sass_file_path_part = $matches[4];

      if (isset($parent_themes_sass_files[$sass_file_context]) && !empty($sass_file_path_part)) {
        // Append scss extension.
        if (strpos($sass_file_path_part, '.scss') === FALSE) {
          $sass_file_path_part .= '.scss';
        }

        $sass_file_path = DRUPAL_ROOT . '/' . $current_themes[$sass_file_context]->subpath . $sass_file_path_part;

        if (isset($parent_themes_sass_files[$sass_file_context][$sass_file_path])) {
          // If this @import partial directive has '_' in the path file name
          // then this is a parent partial and in this case just c&p the line.
          $sass_file_path_info = pathinfo($sass_file_path);
          if (strpos($sass_file_path_info['basename'], '_') === 0) {
            $flatten_lines[] = $line;
          }
          // If this is not a parent partial then lets import it.
          else {
            $new_flattened_files = _bs_base_flatten_sass_file_imports($parent_themes_sass_files[$sass_file_context][$sass_file_path], $sass_file_context, $current_themes, $parent_themes_sass_files, $depth + 1);
            $flatten_lines = _bs_base_array_unique_lines(array_merge($flatten_lines, $new_flattened_files));
          }
        }
        else {
          // If file does not exist check that it is not partial maybe.
          $sass_partial_file_path_part = explode('/', $sass_file_path_part);
          $sass_partial_file_path_part[count($sass_partial_file_path_part) - 1] = '_' . $sass_partial_file_path_part[count($sass_partial_file_path_part) - 1];
          $sass_partial_file_path_part = implode('/', $sass_partial_file_path_part);
          $sass_file_path = DRUPAL_ROOT . '/' . $current_themes[$sass_file_context]->subpath . $sass_partial_file_path_part;

          if (isset($parent_themes_sass_files[$sass_file_context][$sass_file_path])) {
            // @todo - Check for duplicates including commented duplicate for
            // $depth > 0?
            $flatten_lines[] = $line;
          }
        }
      }
      else {
        if ($depth > 0) {
          // Skip @import "init"; lines from parent themes.
          if ($sass_file_context === 'init' && empty($sass_file_path)) {
            continue;
          }
          // Expand the import for all partials from this themes.
          elseif ($sass_file_context === 'partials') {
            // Get the sub-folder sass part of the target file.
            $subfolder = NULL;
            $parts = explode($current_themes[$target_machine_name]->subpath, $target_sass_file->uri);
            if (!empty($parts)) {
              $parts = explode('/', $parts[1]);
              array_pop($parts);
              $subfolder = implode('/', $parts);
            }

            if (empty($subfolder)) {
              drush_log("Sub folder is empty and it is needed for $line.", LogLevel::ERROR);
              continue;
            }
            $line = ($sass_file_commented ? '//' : '') . "@import \"{$target_machine_name}{$subfolder}/{$sass_file_context}{$sass_file_path_part}\";\n";
          }
        }

        // @todo - Check for duplicates including commented duplicate?
        $flatten_lines[] = $line;
      }
    }
    else {
      // If we land here and we are target theme ($depth == 0) then something
      // is probably wrong lets report it.
      // If we are in some parent theme ($depth > 0) then we ignore this line.
      if ($depth === 0) {
        drush_log("Parsing of SASS file $target_sass_file->uri failed for some reason for line  $line\n", LogLevel::ERROR);
      }
    }
  }

  return $flatten_lines;
}

/**
 * Generate child SASS file from parent theme.
 *
 * @param object $parent_sass_file
 *   Plain PHP object holding SASS file information.
 * @param array $options
 *   Array of parent/child options.
 *
 * @return bool
 *   TRUE on success, FALSE other way.
 */
function _bs_base_generate_sass_file($parent_sass_file, array $options) {
  $filepath = $options['child_path'] . '/' . $parent_sass_file->filepath;
  _bs_base_ensure_directory($parent_sass_file->filepath, $options['child_path']);

  $sass_import_path = $options['parent_machine_name'] . '/' . $parent_sass_file->filepath;
  $sass_import_path = substr($sass_import_path, 0, strpos($sass_import_path, '.scss'));

  if (!file_put_contents($filepath, "@import \"init\";\n@import \"{$sass_import_path}\";\n")) {
    drush_log("Failed to generate default SASS file in $filepath.", LogLevel::ERROR);
    return FALSE;
  }
  return $sass_import_path;
}

/**
 * Generate theme file with a default content.
 *
 * @param string $file_name
 *   File name.
 * @param array $options
 *   Options array.
 *
 * @return bool
 *   TRUE on success, FALSE on failure.
 */
function _bs_base_generate_file($file_name, array $options) {
  $content = '';
  switch ($file_name) {
    case $options['child_machine_name'] . '.info.yml':
      // Lets start from parent info file.
      $all_themes = _bs_base_drupal_theme_list_info();
      $yaml_array = $all_themes[$options['parent_machine_name']]->info;

      // Change basic stuff.
      $yaml_array['name'] = $options['child_name'];
      $yaml_array['description'] = $options['child_description'];
      $yaml_array['base theme'] = $options['parent_machine_name'];

      // Override libraries.
      $yaml_array['libraries'] = [$options['child_machine_name'] . '/global-styling'];

      // Remove version, project and datestamp keys that are coming from
      // Drupal.org packager.
      unset($yaml_array['version']);
      unset($yaml_array['project']);
      unset($yaml_array['datestamp']);

      // Remove libraries-extend and component-libraries.
      unset($yaml_array['libraries-extend']);
      unset($yaml_array['component-libraries']);

      $yaml_array['libraries-override'] = _bs_base_generate_libraries_override($options['parent_machine_name']);

      // Finally lets add empty line separators for better visual code grouping.
      $content .= _bs_base_regexp(Yaml::encode($yaml_array), [
        "^regions:" => "\nregions:",
        "^libraries:" => "\nlibraries:",
        "^libraries-override:" => "\nlibraries-override:",
        "^bs_versions:" => "\nbs_versions:",
      ]);
      break;

    case $options['child_machine_name'] . '.libraries.yml':
      // Lets start from parent libraries file.
      $file_content = file_get_contents($options['parent_path'] . '/' . $options['parent_machine_name'] . '.libraries.yml');
      $yaml_array = Yaml::decode($file_content);
      $yaml_array = ['global-styling' => $yaml_array['global-styling']];
      $content .= Yaml::encode($yaml_array);
      break;

    case 'sass/variables/_' . $options['child_machine_name'] . '.scss':
      $content .= "// Variable overrides and custom variable definitions.\n// Note that first all variables from theme-options.yml file will be loaded and\n// then variables from this file.\n";
      break;

    case 'sass/_init.scss':
      $content .= <<<EOD
// Main base file which is responsible of loading all variables in correct
// order and all mixin files we are using.
//
// NOTE that this order is not fixed, feel free to change it you see it fit for
// your custom theme.

// Theme custom variables and overrides.
@import "variables/{$options['child_machine_name']}";

// Load variables and other init code from all parent themes.
@import "{$options['parent_machine_name']}/sass/init";

EOD;
      break;

    case 'gulp-options.yml':
      $content .= "# This file holds various gulp configurations that we need in our Gulp process.\n#\n# Note that options from this file will be merged with Gulp options from parent\n# theme.\n\n";

      // Get parent themes info.
      $content .= "parentTheme:\n  # Order is important and needs to goes from top most parent to bottom.\n  # Default path values will work for standard positions of contrib and custom\n  # themes.\n  # Theme path to parent theme folder can be relative or absolute.";

      $all_themes = _bs_base_drupal_theme_list_info();
      $parent_themes = _bs_base_drupal_get_base_themes($all_themes, $options['parent_machine_name']);
      foreach (array_keys(array_reverse($parent_themes)) as $parent_theme) {
        $relative_path = '../' . UrlGenerator::getRelativePath($options['child_path'], $all_themes[$parent_theme]->subpath) . '/';
        $content .= "\n  -\n    name: '$parent_theme'\n    path: '$relative_path'";
      }

      // Stylintrc file we always reuse from root parent theme.
      $content .= "\n\n# Stylelint config we will reuse from $parent_theme parent theme.\nstylelint:\n  configFile: '$relative_path.stylelintrc.yml'\n";
      break;

    case 'theme-options.yml':
      $content .= <<<EOD
# {$options['child_name']} options.

sass:
  injectVariables:
    variables:
      # Colors

      # General named colors.
      white:      '#fff'
      black:      '#000'
      gray-light: '#868e96'

      # Project specific named colors.
      athens-grey:    '#e2e2e2'
      blue:           '#007bff'
      bleached-cedar: '#4d4d4d'
      cod-grey:       '#292b2c'

      # Theme colors override.
      primary:    '\$blue'
      secondary:  '\$gray-light'

      theme-colors:
        primary: '\$primary'
        secondary: '\$secondary'

      # Color sets overrides.
      colors-default:
        bg:               '\$white'
        color:            '\$cod-grey'
        secondary-color:  '\$bleached-cedar'
        inverted-color:   '\$white'
      colors-inverse:
        bg:               'map-get(\$colors-default, "color")'
        color:            'map-get(\$colors-default, "inverted-color")'
        secondary-color:  'darken(map-get(\$colors-default, "inverted-color"), 17%)'
        inverted-color:   'map-get(\$colors-default, "color")'
      colors-alternate:
        bg:               '\$athens-grey'


      # Typography

      #base-fonts:
      #font-family-base: ''
      #hero-font-family: ''


      # Options

      enable-box-shadow:         true
      enable-box-shadow-hover:   true
      enable-raise-bottom-line:  true


      # Layout

      #brand-logo-width:   200px
      #grid-gutter-width:  24px
      #spacer:             0.75rem


      # Components

      #header-bg:    ''
      #header-color: ''
      #header-border-color: ''

      teaser-primary:
        bg:   '\$primary'
      teaser-secondary:
        bg:   '\$secondary'

EOD;
      break;

    case "config/schema/{$options['child_machine_name']}.schema.yml":
      $content .= Yaml::encode([
        $options['child_machine_name'] . '.settings' => [
          'type' => $options['parent_machine_name'] . '.settings',
          'label' => $options['child_name'] . ' settings',
        ],
      ]);
      break;

    case 'README.md':
      $content .= "THEME DOCUMENTATION\n-------------------\n\nPut your custom theme documentation here.";
      break;

    default:
      drush_log("Generation of file $file_name not supported.", LogLevel::ERROR);
      return FALSE;
  }

  _bs_base_ensure_directory($file_name, $options['child_path']);

  if (!file_put_contents($options['child_path'] . '/' . $file_name, $content)) {
    drush_log("Failed to generate file $file_name.", LogLevel::ERROR);
    return FALSE;
  }

  return TRUE;
}

/**
 * Get all CSS libraries from passed theme that child theme needs to override.
 *
 * @param string $parent_machine_name
 *   Parent theme machine name.
 *
 * @return array
 *   Libraries override array.
 */
function _bs_base_generate_libraries_override($parent_machine_name) {
  $parent_theme_info = _bs_base_get_theme_info($parent_machine_name);

  // Remove all libraries-override elements with false - we don't need them
  // duplicated here.
  $libraries_override = array_filter($parent_theme_info->info['libraries-override']);

  // Change override keys until core improve this in
  // https://www.drupal.org/node/2642122.
  $parent_root_path = '/' . $parent_theme_info->subpath . '/';
  foreach ($libraries_override as $key => &$library) {
    if (isset($library['css'])) {
      foreach ($library['css'] as $library_key => $library_value) {
        $css = reset($library_value);
        // If this is false and the only thing in the library then we should
        // remove it in the same way as we are removing duplicates in
        // previous array_filter.
        if ($css === FALSE && count($library) === 1) {
          unset($libraries_override[$key]);
        }
        else {
          $library['css'][$library_key] = [$parent_root_path . $css => $css];
        }
      }
    }
  }

  // Disable loading of global styles from parent theme.
  $libraries_override[$parent_machine_name . '/global-styling'] = FALSE;

  // Override libraries defined in first parent theme.
  $parent_library_overrides = _bs_base_get_css_libraries_for_override($parent_machine_name);
  foreach ($parent_library_overrides as $key => $parent_library_override) {
    $libraries_override[$key] = $parent_library_override;
  }

  return $libraries_override;
}

/**
 * Get CSS libraries from theme in libraries override format.
 *
 * @param string $theme_machine_name
 *   Theme machine name from which we are getting CSS libraries.
 *
 * @return array
 *   Arrays of CSS libraries in libraries override format.
 */
function _bs_base_get_css_libraries_for_override($theme_machine_name) {
  $theme_info = _bs_base_get_theme_info($theme_machine_name);

  // Check and override any CSS library that is defined in parent theme.
  // @todo - we should allow library override addition only in the case when
  // css files does belong into the flatten SASS/CSS file structure.
  $file_content = file_get_contents($theme_info->subpath . '/' . $theme_machine_name . '.libraries.yml');
  $parent_libraries = Yaml::decode($file_content);

  $parent_library_overrides = [];
  foreach ($parent_libraries as $key => $parent_library) {
    if ($key == 'global-styling') {
      continue;
    }
    if (isset($parent_library['css'])) {
      foreach ($parent_library['css'] as $css_key => $css_library) {
        $css_file = key($css_library);
        $parent_library['css'][$css_key] = [$css_file => $css_file];
      }
      $parent_library_overrides[$theme_machine_name . '/' . $key] = ['css' => $parent_library['css']];
    }
  }

  return $parent_library_overrides;
}

/**
 * Get array of all SASS files in the given path.
 *
 * @param string $path
 *   Theme path.
 *
 * @return array
 *   Array of all SASS files in the given path.
 */
function _bs_base_get_sass_files($path) {
  $files = _bs_base_file_scan_directory(DRUPAL_ROOT . '/' . $path . '/sass', '/.*\.scss$/');

  // Add some more info's that we need.
  foreach ($files as &$file) {
    // Add SASS partial flag.
    $file->partial = strpos($file->name, '_') === 0 ? TRUE : FALSE;

    // Add theme relative file path.
    $file->filepath = substr($file->uri, strpos($file->uri, '/sass/') + 1);
  }

  return $files;
}

/**
 * Regular expression search and replace in the text.
 *
 * @param string $text
 *   Text to search and replace.
 * @param array $regexps
 *   Array of regexps searches with it replace values.
 * @param string $modifiers
 *   PHP regular expression modifiers.
 * @param string $delimiter
 *   PHP regular expression delimiter.
 *
 * @return string
 *   Replaced text.
 */
function _bs_base_regexp($text, array $regexps, $modifiers = 'm', $delimiter = '%') {
  $new_content = $text;
  foreach ($regexps as $pattern => $value) {
    if ($replaced = preg_replace($delimiter . $pattern . $delimiter . $modifiers, $value, $new_content)) {
      $new_content = $replaced;
    }
  }

  return $new_content;
}

/**
 * Regular expression search and replace in the file.
 *
 * @param string $file_name
 *   File path.
 * @param array $regexps
 *   Array of regexps searches with it replace values.
 *
 * @return bool
 *   TRUE on success, FALSE other way.
 */
function _bs_base_regexp_file($file_name, array $regexps) {
  $file_contents = file_get_contents($file_name);
  if ($file_contents === FALSE) {
    drush_log("Can not open file $file_name for regexp search&replace.", LogLevel::WARNING);
    return FALSE;
  }

  return file_put_contents($file_name, _bs_base_regexp($file_contents, $regexps));
}

/**
 * Set primitive values in a yml file.
 *
 * Please note that this implementation is not perfect but exist only to support
 * the needs of this drush implementation. There are couple of limitations that
 * are explained in function comments. Most importantly values in values array
 * can be only primitive types for now.
 *
 * @param string $path
 *   Yaml file path.
 * @param array $values
 *   Array with yaml values to set. The element key is a combination of yaml
 *     keys and element value is yaml value. For example:
 *
 *     $values['logo:path'] = 'custom/logo/path'
 *
 *   will do
 *
 *     logo:
 *       path: 'custom/logo/path'.
 * @param bool $add
 *   If TRUE then if value does not exist it will be added.
 *
 * @throws Exception
 *   Throws exception in the case that we try to set non scalar value.
 */
function _bs_base_set_yml_value($path, array $values, $add = FALSE) {
  $write = FALSE;
  $file_contents = file_get_contents($path);

  foreach ($values as $yml_key => $value) {
    if (!is_scalar($value)) {
      throw new Exception("Can not set non scalar value for $yml_key in $path");
    }

    // Regular expression pattern that can locate and change value based on yaml
    // array keys.
    $pattern = "#" . implode(":\n(?:.|\n)*?", explode('.', $yml_key)) . ":\s(.*?)\n#";

    $count = 0;
    $res = preg_replace_callback($pattern, function ($matches) use ($value) {
      return str_replace($matches[1], $value, $matches[0]);
    }, $file_contents, 1, $count);

    // If variable does not exist and $add flag is turn on we will add this
    // value.
    // @todo, @note - not sure how this will support two call like in the case
    // when bs_versions key does not exist at all:
    //
    // $values = [
    //   'bs_versions.bs_base' => 8000,
    //   'bs_versions.bs_bootstrap' => 8000,
    // ];
    // _bs_base_set_yml_value($path, $values, TRUE);
    //
    // This call will fail in this case. If this happens we will need to
    // fix/improve this code.
    if ($count === 0 && $add) {
      // Convert variable to yaml text format.
      $keys = explode('.', $yml_key);
      $yaml_value = [array_pop($keys) => $value];
      while ($key = array_pop($keys)) {
        $yaml_value = [$key => $yaml_value];
      }

      // Check does part of the yaml value already exist and if yes merge it
      // with a new value.
      // @note - this will use Yaml decode and encode and we will lose any
      // comment that exist. Symfony it self will not implement comments
      // support. @see https://github.com/symfony/symfony/issues/22516.
      // @todo - however Acquia is using consolidation/comments library to
      //   overcome this limitation of Symfony Yaml parser. We could implement
      //   this approach in the future because it would simplify other parts
      //   of drush script.
      //   @see https://github.com/acquia/blt/pull/3629/files.
      //   @see https://github.com/consolidation/comments
      $root_key = key($yaml_value);
      // Regexp to select all indented lines for a given root key.
      // @see https://stackoverflow.com/a/48313919 for a bit more explanation on
      // this pattern (multi line mode variation).
      $root_key_pattern = "/^$root_key:\R(^ +.+\R)+/m";
      $root_key_count = 0;
      $res = preg_replace_callback($root_key_pattern, function ($matches) use ($root_key, $yaml_value) {
        $root_key_value = Yaml::decode($matches[0]);
        $root_key_value = array_merge_recursive($root_key_value, $yaml_value);
        return Yaml::encode($root_key_value);
      }, $file_contents, 1, $root_key_count);
      if ($root_key_count === 1) {
        $file_contents = $res;
      }
      // If value does not exist let us simply add it to the end of the file.
      else {
        $file_contents .= "\n" . Yaml::encode($yaml_value);
      }

      $write = TRUE;
    }
    elseif (!empty($res) && $res !== $file_contents) {
      $file_contents = $res;
      $write = TRUE;
    }
  }

  if ($write) {
    file_put_contents($path, $file_contents);
  }
}

/**
 * Update SASS files in theme and do SASS import flattening.
 *
 * @param string $theme_machine_name
 *   Theme machine name.
 */
function _bs_base_update_sass_files($theme_machine_name) {
  $all_themes = _bs_base_drupal_theme_list_info();
  $parent_themes = _bs_base_get_parent_themes($theme_machine_name);
  $target_path = _bs_base_drupal_get_theme_path($theme_machine_name);
  $first_parent_machine_name = $all_themes[$theme_machine_name]->info['base theme'];

  $options = [
    'parent_machine_name' => $first_parent_machine_name,
    'parent_path' => $all_themes[$first_parent_machine_name]->subpath,
    'child_machine_name' => $theme_machine_name,
    'child_path' => $target_path,
    'child_name' => $all_themes[$theme_machine_name]->info['name'],
    'child_description' => $all_themes[$theme_machine_name]->info['description'],
  ];

  // Build SASS info array of parent themes.
  $parent_themes_sass_files = [];
  /** @var \Drupal\Core\Extension\Extension $theme */
  foreach (array_keys($parent_themes) as $parent_theme_machine_name) {
    $parent_themes_sass_files[$parent_theme_machine_name] = _bs_base_get_sass_files($all_themes[$parent_theme_machine_name]->subpath);
  }

  // Get target SASS files.
  $target_theme_sass_files = _bs_base_get_sass_files($all_themes[$theme_machine_name]->subpath);

  // Add any missing main SASS files from first parent theme to a target theme.
  $new_files = FALSE;
  $new_sass_files = [];
  foreach ($parent_themes_sass_files[$first_parent_machine_name] as $sass_file) {
    $target_theme_sass_uri = DRUPAL_ROOT . '/' . $target_path . '/' . $sass_file->filepath;
    if (!$sass_file->partial && empty($target_theme_sass_files[$target_theme_sass_uri])) {
      $new_sass_files[] = _bs_base_generate_sass_file($sass_file, $options);
      $new_files = TRUE;
    }
  }

  // Make sure that _init and variables/_target_theme_machine_name.scss exists.
  foreach (['sass/variables/_' . $theme_machine_name . '.scss', 'sass/_init.scss'] as $file) {
    if (empty($target_theme_sass_files[DRUPAL_ROOT . '/' . $target_path . '/' . $file])) {
      if (!_bs_base_generate_file($file, $options)) {
        drush_log("Failed to generate default SASS file $file file.", LogLevel::ERROR);
      }
      $new_files = TRUE;
    }
  }

  // If we have new SASS files lets rescan target theme again.
  if ($new_files) {
    $target_theme_sass_files = _bs_base_get_sass_files($target_path);
  }

  // Iterate over all *.scss files and for all non partial files flatten SASS
  // imports.
  foreach ($target_theme_sass_files as $sass_file) {
    // Do not process partials.
    if ($sass_file->partial) {
      continue;
    }

    $flattened_sass = _bs_base_flatten_sass_file_imports($sass_file, $theme_machine_name, $all_themes, $parent_themes_sass_files);

    // Check that parent files partials that are added to new files maybe exist
    // in existing imports. If yes remove them. This is a case when some
    // partials from parent themes are moved to new SASS file due to
    // refactoring.
    foreach ($new_sass_files as $new_file) {
      if (empty($new_file)) {
        continue;
      }

      $new_file_path = $all_themes[$first_parent_machine_name]->parentPath . '/' . $new_file . '.scss';
      $new_sass_file = $parent_themes_sass_files[$first_parent_machine_name][$new_file_path];

      // Skip newly added SASS file.
      if ($new_sass_file->filepath === $sass_file->filepath) {
        continue;
      }

      $new_sass_file_flattened = _bs_base_flatten_sass_file_imports($new_sass_file, $theme_machine_name, $all_themes, $parent_themes_sass_files);
      // Remove first @import "init";
      array_shift($new_sass_file_flattened);

      $remove_lines = array_intersect($flattened_sass, $new_sass_file_flattened);
      if (!empty($remove_lines)) {
        $flattened_sass = array_diff($flattened_sass, $remove_lines);
      }
    }

    // Save the file.
    if (file_put_contents($sass_file->uri, implode('', $flattened_sass)) === FALSE) {
      drush_log("Failed to write {$sass_file->uri} file.", LogLevel::ERROR);
    }
  }
}
